
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>


/**********************************************************************
 * Data Types
 **********************************************************************/
typedef enum  {
  go_empty, go_b, go_w
} Go;


typedef struct  {
  int  prev, x, y, message, end;
} Node;


/**********************************************************************
 * Forward Declarations
 **********************************************************************/
static FILE  *probFile(const char *baseName, int probNum, int *probLen);
static void  printBoard(Go board[19*19], int max);
static const char  *getMessage(int msgVal, int bestResult);
static void  addNode(Node *nodes, int numNodes, int prevNode, Go whoseMove,
		     int bestResult);
static void  addLabels(Node *nodes, int numNodes, int prevNode);


/**********************************************************************
 * Functions
 **********************************************************************/
int  main(int argc, char *argv[])  {
  Go  board[19*19];
  Node  nodes[1000];
  int  whichProb, i, loc, max, curNode;
  FILE  *fIn;
  int  probLen;
  unsigned char  c;
  int  numNodes;
  int  bestResult;

  assert(argc >= 3);
  for (i = 0;  i < 19*19;  ++i)  {
    board[i] = go_empty;
  }

  printf("(;GM[1]FF[3]\n"
	  "SZ[19]HA[0]KM[0]GN[Problem %s:%d]\n", argv[1],
	  atoi(argv[2]));
  for (whichProb = 2;  whichProb < argc;  ++whichProb)  {
    if (argc > 3)
      printf("(;\n");
    printf("AB");

    fIn = probFile(argv[1], atoi(argv[whichProb]), &probLen);
    loc = -1;
    while ((c = getc(fIn)) != 0)  {
      loc += c;
      board[loc] = go_b;
      printf("[%c%c]", (loc%19) + 'a', (loc/19) + 'a');
      --probLen;
    }
    --probLen;
    max = loc + 19;

    loc = -1;
    printf("\nAW");
    while ((c = getc(fIn)) != 0)  {
      --probLen;
      loc += c;
      board[loc] = go_w;
      printf("[%c%c]", (loc%19) + 'a', (loc/19) + 'a');
    }
    printf("\n");
    --probLen;
    if (loc + 19 > max)
      max = loc + 19;

    printBoard(board, max);
    c = getc(fIn);
    /* assert(c == 141); */
    printf("C[Problem set %s, number %d.\n", argv[1], atoi(argv[whichProb]));
    c = getc(fIn);
    /*
     * Decoding the problem description:
     *   42 = 0x2a = Elementary, b to live
     *   52 = 0x34 = Intermediate, b to live
     *   62 = 0x3e = Advanced, b to live
     *   74 = 0x4a = Elementary, b to kill
     *   84 = 0x54 = Intermediate, b to kill
     *   94 = 0x5e = Advanced, b to kill
     *  126 = 0x7e = Advanced, b to live
     */
    switch(c)  {

    case 0x2a:
      printf("Elementary.\nBlack to play and live.]\n");
      break;
    case 0x4a:
      printf("Elementary.\nBlack to play and kill.]\n");
      break;
    case 0x6a:
      printf("Elementary.\nBlack to play and win the capture race.]\n");
      break;

    case 0x34:
      printf("Intermediate.\nBlack to play and live.]\n");
      break;
    case 0x54:
      printf("Intermediate.\nBlack to play and kill.]\n");
      break;
    case 0x74:
      printf("Intermediate.\nBlack to play and win the capture race.]\n");
      break;

    case 0x3e:
      printf("Advanced.\nBlack to play and live.]\n");
      break;
    case 0x5e:
      printf("Advanced.\nBlack to play and kill.]\n");
      break;
    case 0x7e:
      printf("Advanced.\nBlack to play and win the capture race.]\n");
      break;

    default:
      printf("Unknown description code %d (0x%x).]\n", c, c);
      break;
    }
    numNodes = getc(fIn);
    c = getc(fIn);
    numNodes += c * 256;
    assert(numNodes < sizeof(nodes) / sizeof(*nodes));
    for (curNode = 0;  curNode < numNodes;  ++curNode)  {
      unsigned char  nodeIn[5];
      fread(nodeIn, 5, 1, fIn);
      nodes[curNode].prev = ((int)nodeIn[0] | ((int)nodeIn[1] << 8)) & 0x7ff;
      nodes[curNode].x = nodeIn[1] >> 3;
      nodes[curNode].y = nodeIn[2];
      nodes[curNode].message = nodeIn[3];
      nodes[curNode].end = nodeIn[4];
    }
    bestResult = 3;  /* Wrong.  This will change. */
    for (i = 0;  i < numNodes;  ++i)  {
      if (nodes[i].message != 0)  {
	if (((nodes[i].message >> 4) & 7) < bestResult)
	  bestResult = ((nodes[i].message >> 4) & 7);
      }
    }
    addNode(nodes, numNodes, 2047, go_b, bestResult);
    if (argc > 3)
      printf(")");
    fclose(fIn);
  }
  printf(")\n");
  exit(0);
}


static FILE  *probFile(const char *baseName, int probNum, int *probLen)  {
  char  tableFileName[100], problemFileName[100], sgfFileName[100];
  FILE  *fIn;
  int  probStart, probEnd, newLocation;

  assert(probNum > 0);
  assert(strlen(baseName) + 15 <= sizeof(tableFileName));
  sprintf(tableFileName, "%s.dr", baseName);
  sprintf(problemFileName, "%s.dt", baseName);
  sprintf(sgfFileName, "%s-%d.sgf", baseName, probNum);
  fIn = fopen(tableFileName, "r");
  assert(fIn != NULL);
  if (probNum == 1)  {
    probStart = 0;
  } else  {
    newLocation = fseek(fIn, (probNum - 2) * sizeof(int), SEEK_SET);
    if (newLocation < 0) {
      fprintf(stderr, "Sorry, there are only %d problems!\n",
	      (newLocation / sizeof(int)) - 1);
      exit(1);
    }
    fread(&probStart, sizeof(probStart), 1, fIn);
  }
  fread(&probEnd, sizeof(probEnd), 1, fIn);
  fclose(fIn);
  fIn = fopen(problemFileName, "r");
  assert(fIn != NULL);
  fseek(fIn, probStart, SEEK_SET);
  *probLen = probEnd - probStart;
  return(fIn);
}


static void  printBoard(Go board[19*19], int max)  {
  /*
  int  x, y, loc;

  loc = 0;
  for (y = 0;  y < 19;  ++y)  {
    for (x = 0;  x < 19;  ++x)  {
      printf("%c ", ".#O"[board[loc]]);
      ++loc;
    }
    printf("\n");
    if (loc >= max)
      return;
  }
  */
}


static const char  *getMessage(int msgVal, int bestResult)  {
  static char  unknownBuf[100];

  switch(msgVal >> 4)  {
  case 0:
  case 8:
    if (msgVal == 0)
      return("");
    else  {
      assert(bestResult == 0);
      return("Correct.");
    }
  case 1:
    assert(bestResult <= 1);
    if (bestResult == 1)
      return("Correct; advantageous ko.");
    else
      return("Wrong; advantageous ko.");
  case 2:
    assert(bestResult <= 2);
    if (bestResult == 2)
      return("Correct; disadvantageous ko.");
    else
      return("Wrong; disadvantageous ko.");
  case 3:
    return("Wrong.");
  default:
    sprintf(unknownBuf, "Unknown message %d:%d",
	    msgVal & 15, msgVal >> 4);
    return(unknownBuf);
  }
}


static void  addNode(Node *nodes, int numNodes, int prevNode,
		     Go whoseMove, int bestResult)  {
  int  numMatch, i;

  numMatch = 0;
  for (i = 0;  i < numNodes;  ++i)  {
    if (nodes[i].prev == prevNode)
      ++numMatch;
  }
  if (numMatch > 0)  {
    if (numMatch > 1)
      addLabels(nodes, numNodes, prevNode);
    for (i = 0;  i < numNodes;  ++i)  {
      if (nodes[i].prev == prevNode)  {
	if (numMatch > 1)
	  printf("(\n");
	printf(";%c[%c%c]CR[%c%c]\n", "?BW"[whoseMove],
	       nodes[i].x + 'a', nodes[i].y + 'a',
	       nodes[i].x + 'a', nodes[i].y + 'a');
	if (nodes[i].message & 15)  {
	  printf("C[%s]\n", getMessage(nodes[i].message, bestResult));
	} else  {
	  assert(nodes[i].message == 0);
	}
	addNode(nodes, numNodes, i, whoseMove ^ (go_b ^ go_w), bestResult);
	if (numMatch > 1)
	  printf(")\n");
      }
    }
  }
}


static void  addLabels(Node *nodes, int numNodes, int prevNode)  {
  int  i;
  char  label = 'A';

  printf(";LB");
  for (i = 0;  i < numNodes;  ++i)  {
    if (nodes[i].prev == prevNode)  {
      printf("[%c%c:%c]",
	      nodes[i].x + 'a',
	      nodes[i].y + 'a',
	      label++);
    }
  }
  printf("\n");
}
